import os
import stat

def read_int(filename):
    """Read integer from file filename"""
    with open(filename) as file_id:
        return int(file_id.readline())


class Lperf:
    'Monitoring using a modified version of Linux Perf tool'
    def __init__(self, sensor_set={'instructions'}, interval=100):

        self.executor = None
        self.names = sensor_set
        self.interval = interval
        self.cmdline = ''
        
    def build(self, executor):
        """Installs and modifies the Linux Perf Tool"""

        current_path=os.path.dirname(os.path.abspath(__file__))
        if not os.path.exists('/tmp/lperf/tools/perf/perf'):
            executor.local('mkdir /tmp/lperf; tar xj -C /tmp/lperf/ -f %s/lperf_tools.tar.bz2' % current_path)
            executor.local('patch -i %s/lperf_builtin-stat.diff /tmp/lperf/tools/perf/builtin-stat.c' % current_path)
            executor.local('cd /tmp/lperf/tools/perf/; make')
        executor.local('cp /tmp/lperf/tools/perf/perf /tmp/bin/lperf')
        if read_int('/proc/sys/kernel/perf_event_paranoid') != 0:
            executor.hosts("sh -c 'echo 0 >/proc/sys/kernel/perf_event_paranoid'", root=True)

        self.executor = executor

        self.cmdline = '/tmp/bin/lperf stat -x " " -I %s -a -A -e %s' % (self.interval, ' '.join(self.names))
        self.cmdline += ' -o /dev/shm/lperf_monitoring &'

    def start(self):
        'Starts the monitoring right before the benchmark'
        self.executor.hosts(self.cmdline)

    def stop(self):
        'Stops the monitoring right before the benchmark'
        self.executor.hosts('killall lperf')

    def save(self, experiment, benchname, beg_time):
        'Save the results when time is no more critical'
        filename_moj = experiment.output_file+'_lperf'
        os.makedirs(filename_moj, exist_ok=True)
        if len(self.executor.hostnames) > 1:
            for hostname in self.executor.hostnames:
                self.executor.local('oarcp %s:/dev/shm/lperf_monitoring %s/%s_%s_%s' %
                                    (hostname, filename_moj, hostname, benchname, beg_time))
        else:
            self.executor.local('cp /dev/shm/lperf_monitoring %s/%s_%s_%s' %
                                    (filename_moj, 'localhost', benchname, beg_time))



import time

class SleepBench:
    'Sleep benchmark'
    def __init__(self, default_time=30):
        'The only mandatory filed is names'
        self.names = {'sleep'}
        self.time = default_time

    def build(self, executor):
        'returns the parameters'
        return {'sleep':[self.time]}

    def run(self, bench, params, executor):
        'Runs the sleep benchmark'
        time.sleep(params)
        return str(params), 'sleep-'+str(params)



import json
import os

###### ! to move to executor
import execo

from expetator.leverages import dvfs
from expetator.monitors import mojitos

def save_scheduler(code_elements, skel_path, output_path, debug=False):
    name, arguments, nb_arguments, pcts, code = code_elements

    with open(skel_path) as file:
        content = file.read()

    content = content.replace('//DEFINE PCT', pcts)
    content = content.replace('ARGUMENTS;', arguments, 1)
    content = content.replace('NBARGUMENTS;', nb_arguments, 1)
    content = content.replace('//CORE_LOOP', code)

    if debug:
        print(content)
    with open(output_path, 'w') as file:
        file.write(content)    

class NeoSched(dvfs.Dvfs):
    def __init__(self, schedulers, dummy=False, baseline=False):
        dvfs.Dvfs.__init__(self, dummy=dummy, baseline=baseline)
        self.schedulers = {}

        for file_sched in schedulers:
            with open(file_sched) as file_id:
                element = json.load(file_id)
                self.schedulers[element[0]] = element

        self.names = list(self.schedulers)
    
    def build(self, executor):
        dvfs.Dvfs.build(self, executor)
        self.available_frequencies = self.names + self.available_frequencies
        moj = mojitos.Mojitos()
        moj.build(executor)
        basedir = os.path.dirname(os.path.abspath(__file__))
        executor.local('cp -f %s/neosched_* /tmp/mojitos/' % basedir)

        executor.local('mkdir -p /tmp/neosched')
        for scheduler in self.names:
            save_scheduler(self.schedulers[scheduler], '/tmp/mojitos/neosched_skel.c',
                           '/tmp/mojitos/neosched_lib.c')

            executor.local('cd /tmp/mojitos; make -f neosched_mak')
            executor.local('cp -f /tmp/mojitos/neosched /tmp/neosched/%s' % scheduler)

        executor.sync('/tmp/neosched')
        self.algorithm = 0

    def start(self, freq):
        self.algorithm = freq
        if freq in self.schedulers:
            print('starts /tmp/neosched/%s' %freq)
            self.remote=execo.action.Remote('/tmp/neosched/%s' % freq,
                                            self.executor.hostnames,
                                            connection_params={'ssh':self.executor.ssh}).start()
        else:
            dvfs.Dvfs.start(self, freq)

    def stop(self, output_file = None):
        if self.algorithm in self.schedulers:
            execo.action.Remote('killall /tmp/neosched/%s' %
                                self.algorithm,
                                self.executor.hostnames,
                                connection_params={'ssh':self.executor.ssh}).run()
            self.remote.wait()

            if not output_file is None:
                dest = output_file % 'neosched'
                os.makedirs(os.path.dirname(dest), exist_ok=True)
                self.executor.local('mv /dev/shm/neosched.data %s' % dest)

        dvfs.Dvfs.stop(self, output_file)

    def get_state(self):
        if self.algorithm in self.schedulers:
            return (0, self.algorithm)
        else:
            return dvfs.Dvfs.get_state(self)



import time
import datetime
import requests
import os
import json

def get_names():
    return {'kwollect'}

class Power:
    'Monitoring using Kwollect on g5k'
    def __init__(self):
        self.hostnames = None
        self.site = None
        self.start_time = None
        self.end_time = None
        self.executor = None
        self.names = get_names()
        
    def build(self, executor):
        """Prepare the right request only needed on g5k"""
        self.hostnames = ','.join([fullname.split('.')[0] for fullname in executor.hostnames])
        self.site = executor.hostnames[0].split('.')[1]
        self.executor = executor
        
    def start(self):
        'Starts the monitoring right before the benchmark'
        self.start_time = int(time.time())
            
    def stop(self):
        'Stops the monitoring right after the benchmark'
        self.end_time = int(time.time())

    def save(self, experiment, benchname, beg_time):
        'Save the results when time is no more critical'
        filename_power = experiment.output_file+'_power'
        os.makedirs(filename_power, exist_ok=True)
        target =  '%s/%s_%s_%s' % (filename_power, self.executor.hostnames[0], benchname, beg_time)

        request = 'https://api.grid5000.fr/stable/sites/%s/metrics?metrics=wattmetre_power_watt&start_time=%s&end_time=%s&nodes=%s' % (self.site, self.start_time, self.end_time, self.hostnames)

        delta = int(time.time())-self.end_time
        if delta < 6:
            time.sleep(6-delta)

        res = requests.get(request)

        data = res.text
        
        raws = json.loads(data)
        host_list = self.hostnames.split(',')
        
        timestamps = {host:[] for host in host_list}
        values = {host:[] for host in host_list}
        for elements in raws:
            device_id = elements['device_id']
            try:
                timestamps[device_id].append(datetime.datetime.strptime(elements['timestamp'], "%Y-%m-%dT%H:%M:%S%z").timestamp())
            except:
                timestamps[device_id].append(datetime.datetime.strptime(elements['timestamp'], "%Y-%m-%dT%H:%M:%S.%f%z").timestamp())

            values[device_id].append(elements['value'])
        result = [(host, timestamps[host], values[host]) for host in host_list]

        with open(target, 'w') as file_id:
            file_id.write(json.dumps(result))



import time
import requests
import os
import json

def get_names():
    return {'power'}

class Power:
    'Monitoring using Power on g5k'
    def __init__(self):
        self.hostnames = None
        self.site = None
        self.start_time = None
        self.end_time = None
        self.executor = None
        self.names = get_names()
        self.g5k = 'OAR_NODE_FILE' in os.environ
        self.local_script = '%s/laptop_power_monitor.py &' % os.path.dirname(os.path.abspath(__file__))
        
    def build(self, executor):
        """Prepare the right request only needed on g5k"""
        if self.g5k:
            self.hostnames = ','.join([fullname.split('.')[0] for fullname in executor.hostnames])
            self.site = executor.hostnames[0].split('.')[1]
        self.executor = executor
        
    def start(self):
        'Starts the monitoring right before the benchmark'
        if self.g5k:
            self.start_time = int(time.time())
        else:
            self.executor.local(self.local_script)
            
    def stop(self):
        'Stops the monitoring right after the benchmark'
        if self.g5k:
            self.end_time = int(time.time())
        else:
            self.executor.local('killall laptop_power_monitor.py')

    def save(self, experiment, benchname, beg_time):
        'Save the results when time is no more critical'
        filename_power = experiment.output_file+'_power'
        os.makedirs(filename_power, exist_ok=True)
        target =  '%s/%s_%s_%s' % (filename_power, self.executor.hostnames[0], benchname, beg_time)
        if self.g5k:

            request = 'https://api.grid5000.fr/stable/sites/%s/metrics/power/timeseries?resolution=1&only=%s&from=%s&to=%s' % (self.site, self.hostnames, self.start_time, self.end_time)
            time.sleep(5)
            res = requests.get(request)

            
            if res.status_code == 200:
                data = res.text
                raws = json.loads(data)['items']

                result = [(raw['uid'], raw['timestamps'],raw['values']) for raw in raws]
        else:
            self.executor.local('cp /dev/shm/power_measures %s' % target)
            with open('/dev/shm/power_measures') as file_id:
                content = [line.split() for line in file_id.readlines()[1:]]
            content = [(int(t), float(p)) for t,p in content]
            result = [('localhost',)+tuple(zip(*content))]

        with open(target, 'w') as file_id:
            file_id.write(json.dumps(result))



import os
import stat

def read_int(filename):
    """Read integer from file filename"""
    with open(filename) as file_id:
        return int(file_id.readline())


network_names = {'rxb', 'rxp', 'txb', 'txp'}
infiniband_names = {'irxb', 'irxp', 'itxb', 'itxp'}
rapl_names = {'dram0', 'dram1', 'package-00', 'package-11'}
load_names = {'user', 'nice', 'system', 'idle', 'iowait', 'irq',
              'softirq', 'steal', 'guest', 'guest_nice'}
perf_names = {'cpu_cycles', 'instructions', 'cache_references', 'cache_misses',
              'branch_instructions', 'branch_misses', 'bus_cycles', 'ref_cpu_cycles',
              'cache_l1d', 'cache_ll', 'cache_dtlb', 'cache_itlb', 'cache_bpu',
              'cache_node', 'cache_op_read', 'cache_op_prefetch', 'cache_result_access',
              'cpu_clock', 'task_clock', 'page_faults', 'context_switches',
              'cpu_migrations', 'page_faults_min', 'page_faults_maj',
              'alignment_faults', 'emulation_faults', 'dummy', 'bpf_output'}

def get_names():
    return network_names | rapl_names | load_names | perf_names | infiniband_names

def get_structured_names():
    return {'network':network_names,
            'rapl':rapl_names,
            'load':load_names,
            'perfct':perf_names,
            'infiniband': infiniband_names}

class Mojitos:
    'Monitoring using MojitO/S'
    def __init__(self, sensor_set={'dram0'}, frequency=10):

        self.perf = len(sensor_set & perf_names) != 0
        self.network = len(sensor_set & network_names) != 0
        self.rapl = len(sensor_set & rapl_names) != 0
        self.load = len(sensor_set & load_names) != 0    
        self.infiniband = len(sensor_set & infiniband_names) != 0    
        
        self.executor = None

        self.names = sensor_set
        self.frequency = frequency
        self.cmdline = ''
        
    def build(self, executor):
        """Installs the mojito/s monitoring framework and add the permissions"""

        if True or self.rapl:
            # should work but do not work currently as to compile mojitos it
            # is always necessary to have rapl. Todo: update mojitos
            if False in [os.path.exists(filename) for filename in
                         ['/usr/share/doc/powercap-utils',
                          '/usr/share/doc/libpowercap-dev',
                          '/usr/share/doc/libpowercap0']]:
                executor.hosts('apt install -y libpowercap0 libpowercap-dev powercap-utils', root=True)
        if not os.path.exists('/tmp/mojitos'):
            executor.local('cd /tmp; git clone https://git.renater.fr/anonscm/git/mojitos/mojitos.git')
        else:
            executor.local('cd /tmp/mojitos; git pull')
        executor.local('cd /tmp/mojitos; make')
        executor.local('cp /tmp/mojitos/mojitos /tmp/bin/')
        if True or self.rapl:
            if read_int('/proc/sys/kernel/perf_event_paranoid') != 0:
                executor.hosts("sh -c 'echo 0 >/proc/sys/kernel/perf_event_paranoid'", root=True)
            mode = os.stat('/sys/class/powercap/intel-rapl/intel-rapl:0/constraint_0_max_power_uw')
            if not mode.st_mode & stat.S_IWUSR:
                executor.hosts("chmod a+rw /sys/class/powercap/intel-rapl/*/*", root=True)
                executor.hosts("chmod a+rw /sys/class/powercap/intel-rapl/*/*/*", root=True)

        self.executor = executor

        self.cmdline = '/tmp/bin/mojitos -t 0 -f %s' % self.frequency
        if self.perf:
            self.cmdline += ' -p ' + ','.join(self.names & perf_names)
        if self.network:
            self.cmdline += ' -d X'
        if self.infiniband:
            self.cmdline += ' -i X'
        if self.rapl:
            self.cmdline += ' -r'
        if self.load:
            self.cmdline += ' -u'
        self.cmdline += ' -o /dev/shm/monitoring &'


    def start(self):
        'Starts the monitoring right before the benchmark'
        self.executor.hosts(self.cmdline)

    def stop(self):
        'Stops the monitoring right before the benchmark'
        self.executor.hosts('killall mojitos')

    def save(self, experiment, benchname, beg_time):
        'Save the results when time is no more critical'
        filename_moj = experiment.output_file+'_mojitos'
        os.makedirs(filename_moj, exist_ok=True)
        if len(self.executor.hostnames) > 1:
            for hostname in self.executor.hostnames:
                self.executor.local('oarcp %s:/dev/shm/monitoring %s/%s_%s_%s' %
                                    (hostname, filename_moj, hostname, benchname, beg_time))
        else:
            self.executor.local('cp /dev/shm/monitoring %s/%s_%s_%s' %
                                    (filename_moj, 'localhost', benchname, beg_time))



class Template:
    'Template for leverages'
    def __init__(self):
        'Initializes the internal variables using parameters'
        pass

    def build(self, executor):
        'Builds the executable and acquire data on the system'
        self.executor = executor
        pass
    
    def available_states(self):
        'Returns all the possible states as a list'
        return [None]

    def start(self, state):
        'Sets the right state on all hosts'
        pass
    
    def stop(self, output_file=None):
        'Reverts to the standard state on all nodes and saves some info in the right directory'
            
    def get_state(self):
        'Returns the current state'
        return None

    def state_to_str(self):
        'Converts a state in a string'
        return 'None'

    def get_labels(self):
        'Returns the labels for the leverage'
        return (Template)
